@@ -15,13 +15,30 @@ see <https://www.gnu.org/licenses/>.
 import logging
 import pandas as pd
 from bam_filter.sam_utils import process_bam, filter_reference_BAM
-import numpy as np
 from bam_filter.utils import get_arguments, create_output_files
+from bam_filter.entropy import find_knee
 import json
+import warnings
 
 log = logging.getLogger("my_logger")
 
 
+def handle_warning(message, category, filename, lineno, file=None, line=None):
+    print("A warning occurred:")
+    print(message)
+    print("Do you wish to continue?")
+
+    while True:
+        response = input("y/n: ").lower()
+        if response not in {"y", "n"}:
+            print("Not understood.")
+        else:
+            break
+
+    if response == "n":
+        raise category(message)
+
+
 def obj_dict(obj):
     return obj.__dict__
 
@@ -39,14 +56,23 @@ def main():
         logging.DEBUG if args.debug else logging.INFO
     )
 
+    if args.debug:
+        warnings.showwarning = handle_warning
+    else:
+        warnings.filterwarnings("ignore")
+
     out_files = create_output_files(prefix=args.prefix, bam=args.bam)
 
     data = process_bam(
         bam=args.bam,
         threads=args.threads,
         reference_lengths=args.reference_lengths,
-        scale=args.scale,
         min_read_count=args.min_read_count,
+        scale=args.scale,
+        sort_memory=args.sort_memory,
+        plot=args.plot,
+        plots_dir=out_files["coverage_plot_dir"],
+        chunksize=args.chunk_size,
     )
 
     data = list(filter(None, data))
@@ -62,13 +88,34 @@ def main():
     logging.info(f"Writing reference statistics to {out_files['stats']}")
     data_df.to_csv(out_files["stats"], sep="\t", index=False, compression="gzip")
 
+    if args.min_norm_entropy == "auto" or args.min_norm_gini == "auto":
+        if data_df.shape[0] > 1:
+            min_norm_gini, min_norm_entropy = find_knee(
+                data_df, out_plot_name=out_files["knee_plot"]
+            )
+            if min_norm_gini is None:
+                min_norm_gini, min_norm_entropy = (
+                    args.min_norm_gini,
+                    args.min_norm_entropy,
+                )
+        else:
+            min_norm_gini = 0.5
+            min_norm_entropy = 0.75
+            logging.warning(
+                f"There's only one genome. Using min_norm_gini={min_norm_gini} and min_norm_entropy={min_norm_entropy}. Please check the results."
+            )
+    else:
+        min_norm_gini, min_norm_entropy = args.min_norm_gini, args.min_norm_entropy
+
     filter_conditions = {
         "min_read_length": args.min_read_length,
         "min_read_count": args.min_read_count,
         "min_expected_breadth_ratio": args.min_expected_breadth_ratio,
         "min_breadth": args.min_breadth,
-        "min_read_ani": args.min_read_ani,
+        "min_avg_read_ani": args.min_avg_read_ani,
         "min_coverage_evenness": args.min_coverage_evenness,
+        "min_norm_entropy": min_norm_entropy,
+        "min_norm_gini": min_norm_gini,
     }
     if args.only_stats:
         logging.info("Skipping filtering...")
