@@ -1,6 +1,6 @@
 import pysam
 import numpy as np
-import re, os, sys
+import os, sys
 import pandas as pd
 from multiprocessing import Pool
 import functools
@@ -9,13 +9,20 @@ import pysam
 import tqdm
 import logging
 import warnings
-from bam_filter.utils import is_debug, calc_chunksize, fast_flatten, initializer
+from bam_filter.utils import is_debug, calc_chunksize, initializer
+from bam_filter.entropy import entropy, norm_entropy, gini_coeff, norm_gini_coeff
+
 import pyranges as pr
-from collections import defaultdict
+
+# import cProfile as profile
+import pstats
+
+import matplotlib.pyplot as plt
 
 log = logging.getLogger("my_logger")
 
-sys.setrecursionlimit(10 ** 6)
+
+sys.setrecursionlimit(10**6)
 
 # Function to calculate evenness of coverage
 def coverage_evenness(coverage):
@@ -31,12 +38,19 @@ def coverage_evenness(coverage):
     # N = len(X)
     # n = len(D2)
     # E = 1 - (n - sum(D2) / C) / N
-    C = float(round(np.mean(coverage)))
-    D2 = [x for x in coverage if x <= C]
+    C = float(np.rint(np.mean(coverage)))
+    D2 = coverage[coverage <= C]
+    # print(len(D2))
+    # D2 = [x for x in coverage if x <= C]
+    # print(len(D2))
+    # exit()
     if len(D2) == 0:  # pragma: no cover
         covEvenness = 1.0
     else:
-        covEvenness = 1.0 - (len(D2) - sum(D2) / C) / len(coverage)
+        if C > 0:
+            covEvenness = 1.0 - (len(D2) - sum(D2) / C) / len(coverage)
+        else:
+            covEvenness = 0.0
 
     return covEvenness
 
@@ -77,12 +91,21 @@ def create_pyranges(reference, starts, ends, strands):
     )
 
 
-def get_bam_stats(params, ref_lengths=None, scale=1e6):
+def get_bam_stats(
+    params,
+    ref_lengths=None,
+    min_read_ani=90.0,
+    scale=1e6,
+    plot=False,
+    plots_dir="coverage-plots",
+):
     """
     Worker function per chromosome
     loop over a bam file and create tuple with lists containing metrics:
     two definitions of the edit distances to the reference genome scaled by aligned read length
     """
+    # prof = profile.Profile()
+    # prof.enable()
     bam, reference = params
     samfile = pysam.AlignmentFile(bam, "rb")
     edit_distances = []
@@ -110,36 +133,39 @@ def get_bam_stats(params, ref_lengths=None, scale=1e6):
     ends = []
     strands = []
     for aln in samfile.fetch(reference=reference, multiple_iterators=False):
-        n_alns += 1
-        if aln.has_tag("AS"):
-            read_aln_score.append(aln.get_tag("AS"))
-        else:
-            read_aln_score.append(np.nan)
+        ani_read = (1 - ((aln.get_tag("NM") / aln.infer_query_length()))) * 100
+        if ani_read >= min_read_ani:
+            n_alns += 1
+            if aln.has_tag("AS"):
+                read_aln_score.append(aln.get_tag("AS"))
+            else:
+                read_aln_score.append(np.nan)
 
-        if aln.has_tag("AS"):
-            edit_distances.append(aln.get_tag("NM"))
-            ani_nm.append((1 - ((aln.get_tag("NM") / aln.infer_query_length()))) * 100)
-        else:
-            edit_distances.append(np.nan)
-            ani_nm.append(np.nan)
+            if aln.has_tag("AS"):
+                edit_distances.append(aln.get_tag("NM"))
+                ani_nm.append(ani_read)
+            else:
+                edit_distances.append(np.nan)
+                ani_nm.append(np.nan)
 
-        read_gc_content.append(calc_gc_content(aln.query_sequence))
-        read_length.append(aln.infer_read_length())
-        read_aligned_length.append(aln.query_alignment_length)
-        read_mapq.append(aln.mapping_quality)
-        read_names.append(aln.query_name)
-        # check if strand is reverse
-        if aln.is_reverse:
-            strand = "-"
-        else:
-            strand = "+"
-        starts.append(aln.reference_start)
-        ends.append(aln.reference_end)
-        strands.append(strand)
+            read_gc_content.append(calc_gc_content(aln.query_sequence))
+            read_length.append(aln.infer_read_length())
+            read_aligned_length.append(aln.query_alignment_length)
+            read_mapq.append(aln.mapping_quality)
+            read_names.append(aln.query_name)
+            # check if strand is reverse
+            if aln.is_reverse:
+                strand = "-"
+            else:
+                strand = "+"
+            starts.append(aln.reference_start)
+            ends.append(aln.reference_end)
+            strands.append(strand)
 
     # get bases covered by reads pileup
-    cov_pos = [
-        pileupcolumn.n
+
+    cov_pos_raw = [
+        (pileupcolumn.pos, pileupcolumn.n)
         for pileupcolumn in samfile.pileup(
             reference,
             start=None,
@@ -150,14 +176,22 @@ def get_bam_stats(params, ref_lengths=None, scale=1e6):
         )
     ]
     samfile.close()
+
+    cov_pos = [i[1] for i in cov_pos_raw]
     # convert datafrane to pyranges
     ranges = create_pyranges(reference, starts, ends, strands)
-    ranges = ranges.merge(strand=False).lengths().to_list()
+    if ranges.df.shape[0] == 0:
+        log.debug(f"No alignments found for {reference}")
+        return None
+    ranges_raw = ranges.merge(strand=False)
+    ranges = ranges_raw.lengths().to_list()
+
     max_covered_bases = np.max(ranges)
     mean_covered_bases = np.mean(ranges)
     bases_covered = int(len(cov_pos))
     # get SD from covered bases
     cov_sd = np.std(cov_pos, ddof=1)
+    cov_var = np.var(cov_pos, ddof=1)
     # get average coverage
     mean_coverage = sum(cov_pos) / reference_length
     mean_coverage_covered = sum(cov_pos) / bases_covered
@@ -169,14 +203,33 @@ def get_bam_stats(params, ref_lengths=None, scale=1e6):
     if breadth_exp_ratio > 1:
         breadth_exp_ratio = 1.0
 
-    cov_evenness = coverage_evenness(cov_pos)
+    # fill vector with zeros to match reference length
+    cov_pos_zeroes = np.pad(cov_pos, (0, reference_length - len(cov_pos)), "constant")
+    cov_evenness = coverage_evenness(cov_pos_zeroes)
     gc_content = (np.sum(read_gc_content) / np.sum(read_length)) * 100
     c_v = cov_sd / mean_coverage
+    d_i = cov_var / mean_coverage
+
     read_mapq = [np.nan if x == 255 else x for x in read_mapq]
 
     tax_abund_aln = round((n_alns / reference_length) * scale)
     tax_abund_read = round((len(set(read_names)) / reference_length) * scale)
 
+    # Analyse site distribution
+    cov_positions = [i[0] for i in cov_pos_raw]
+    n_sites = len(cov_positions)
+    genome_length = bam_reference_length
+    # Site density (sites per thousand bp)
+    site_density = 1000 * n_sites / genome_length
+
+    counts, bins = np.histogram(cov_positions, bins="auto", range=(0, genome_length))
+    n_bins = len(bins)
+
+    entr = entropy(counts)  # Positional entropy
+    norm_entr = norm_entropy(counts)  # Normalized positional entropy
+    gini = gini_coeff(counts)  # Gini coefficient
+    norm_gini = norm_gini_coeff(counts)  # Normalized Gini coefficient
+
     log.debug(f"Number of reads: {len(set(read_names)):,}")
     log.debug(f"Number of alignments: {n_alns:,}")
     log.debug(f"Bases covered: {bases_covered:,}")
@@ -188,12 +241,46 @@ def get_bam_stats(params, ref_lengths=None, scale=1e6):
     log.debug(f"Breadth: {breadth:.2f}")
     log.debug(f"Exp. breadth: {exp_breadth:.2f}")
     log.debug(f"Breadth/exp. ratio: {breadth_exp_ratio:.2f}")
+    log.debug(f"Number of bins: {n_bins}")
+    log.debug(f"Site density: {site_density:.2f}")
+    log.debug(f"Entropy (H): {entr:.2f}")
+    log.debug(f"Normalized entropy (H*): {norm_entr:.2f}")
+    log.debug(f"Gini coefficient (G): {gini:.2f}")
+    log.debug(f"Normalized Gini coefficient (G*): {norm_gini:.2f}")
     log.debug(f"Cov. evenness: {cov_evenness:.2f}")
     log.debug(f"C_v: {c_v:.2f}")
+    log.debug(f"D_i: {d_i:.2f}")
     log.debug(f"Mean mapq: {np.mean(read_mapq):.2f}")
     log.debug(f"GC content: {gc_content:.2f}")
     log.debug(f"Taxonomic abundance (alns): {tax_abund_aln:.2f}")
     log.debug(f"Taxonomic abundance (reads): {tax_abund_read:.2f}")
+
+    if plot:
+        fig, ax = plt.subplots(nrows=1, ncols=1)  # create figure & 1 axis
+        # infer number of bins using Freedman-Diaconis rule
+        positions_cov_zeros = pd.DataFrame({"pos": range(1, bam_reference_length + 1)})
+        positions_cov = pd.DataFrame(
+            {"pos": [i[0] for i in cov_pos_raw], "cov": [i[1] for i in cov_pos_raw]}
+        )
+        positions_cov = positions_cov_zeros.merge(positions_cov, on="pos", how="left")
+        positions_cov["cov"] = positions_cov["cov"].fillna(0)
+        positions_cov["cov"] = positions_cov["cov"].astype(int)
+        positions_cov["cov_binary"] = positions_cov["cov"].apply(
+            lambda x: 1 if x > 0 else 0
+        )
+        plt.plot(
+            positions_cov["pos"],
+            positions_cov["cov"],
+            color="c",
+            ms=0.5,
+        )
+        plt.suptitle(f"{reference}")
+        plt.title(
+            f"cov:{mean_coverage:.4f} b/e:{breadth_exp_ratio:.2f} cov_e:{cov_evenness:.2f} entropy:{norm_entr:.2f} gini:{norm_gini:.2f}"
+        )
+        fig.savefig(f"{plots_dir}/{reference}_coverage.png", dpi=300)
+        plt.close(fig)
+
     data = BamAlignment(
         reference=reference,
         n_alns=n_alns,
@@ -208,7 +295,14 @@ def get_bam_stats(params, ref_lengths=None, scale=1e6):
         breadth=breadth,
         exp_breadth=exp_breadth,
         breadth_exp_ratio=breadth_exp_ratio,
+        n_bins=n_bins,
+        site_density=site_density,
+        entropy=entr,
+        norm_entropy=norm_entr,
+        gini=gini,
+        norm_gini=norm_gini,
         c_v=c_v,
+        d_i=d_i,
         edit_distances=edit_distances,
         # edit_distances_md=edit_distances_md,
         ani_nm=ani_nm,
@@ -222,6 +316,10 @@ def get_bam_stats(params, ref_lengths=None, scale=1e6):
         tax_abund_aln=tax_abund_aln,
         tax_abund_read=tax_abund_read,
     )
+    # prof.disable()
+    # # print profiling output
+    # stats = pstats.Stats(prof).strip_dirs().sort_stats("tottime")
+    # stats.print_stats(5)  # top 10 rows
     return data
 
 
@@ -252,7 +350,14 @@ class BamAlignment:
         breadth,
         exp_breadth,
         breadth_exp_ratio,
+        n_bins,
+        site_density,
+        entropy,
+        norm_entropy,
+        gini,
+        norm_gini,
         c_v,
+        d_i,
         cov_evenness,
         read_names,
         read_aln_score,
@@ -280,7 +385,14 @@ class BamAlignment:
         self.breadth = breadth
         self.exp_breadth = exp_breadth
         self.breadth_exp_ratio = breadth_exp_ratio
+        self.n_bins = n_bins
+        self.site_density = site_density
+        self.entropy = entropy
+        self.norm_entropy = norm_entropy
+        self.gini = gini
+        self.norm_gini = norm_gini
         self.c_v = c_v
+        self.d_i = d_i
         self.cov_evenness = cov_evenness
         self.read_names = read_names
         self.tax_abund_aln = tax_abund_aln
@@ -310,7 +422,14 @@ class BamAlignment:
             "breadth": self.breadth,
             "exp_breadth": self.exp_breadth,
             "breadth_exp_ratio": self.breadth_exp_ratio,
+            "n_bins": self.n_bins,
+            "site_density": self.site_density,
+            "entropy": self.entropy,
+            "norm_entropy": self.norm_entropy,
+            "gini": self.gini,
+            "norm_gini": self.norm_gini,
             "c_v": self.c_v,
+            "d_i": self.d_i,
             "cov_evenness": self.cov_evenness,
             "tax_abund_read": self.tax_abund_read,
             "tax_abund_aln": self.tax_abund_aln,
@@ -363,7 +482,14 @@ class BamAlignment:
             "breadth": self.breadth,
             "exp_breadth": self.exp_breadth,
             "breadth_exp_ratio": self.breadth_exp_ratio,
+            "n_bins": self.n_bins,
+            "site_density": self.site_density,
+            "entropy": self.entropy,
+            "norm_entropy": self.norm_entropy,
+            "gini": self.gini,
+            "norm_gini": self.norm_gini,
             "c_v": self.c_v,
+            "d_i": self.d_i,
             "cov_evenness": self.cov_evenness,
             "tax_abund_read": self.tax_abund_read,
             "tax_abund_aln": self.tax_abund_aln,
@@ -379,7 +505,18 @@ class BamAlignment:
 
 
 # Inspired from https://gigabaseorgigabyte.wordpress.com/2017/04/14/getting-the-edit-distance-from-a-bam-alignment-a-journey/
-def process_bam(bam, threads=1, reference_lengths=None, scale=1e6, min_read_count=0):
+def process_bam(
+    bam,
+    threads=1,
+    reference_lengths=None,
+    min_read_ani=90.0,
+    min_read_count=10,
+    scale=1e6,
+    sort_memory="1G",
+    plot=False,
+    plots_dir="coverage-plots",
+    chunksize=None,
+):
     """
     Processing function: calls pool of worker functions
     to extract from a bam file two definitions of the edit distances to the reference genome scaled by read length
@@ -410,6 +547,14 @@ def process_bam(bam, threads=1, reference_lengths=None, scale=1e6, min_read_coun
             sys.exit(1)
         max_chr_length = np.max(ref_lengths["length"].tolist())
 
+    # Check if BAM files is not sorted by coordinates, sort it by coordinates
+    if not samfile.header["HD"]["SO"] == "coordinate":
+        log.info("BAM file is not sorted by coordinates, sorting it...")
+        sorted_bam = bam.replace(".bam", ".bf-sorted.bam")
+        pysam.sort("-@", str(threads), "-m", str(sort_memory), "-o", sorted_bam, bam)
+        bam = sorted_bam
+        samfile = pysam.AlignmentFile(bam, "rb")
+
     if not samfile.has_index():
         logging.info(f"BAM index not found. Indexing...")
         if max_chr_length > 536870912:
@@ -426,10 +571,24 @@ def process_bam(bam, threads=1, reference_lengths=None, scale=1e6, min_read_coun
     total_refs = samfile.nreferences
     logging.info(f"Found {total_refs:,} reference sequences")
     # logging.info(f"Found {samfile.mapped:,} alignments")
-
     logging.info(
         f"Removing references without mappings or less than {min_read_count} reads..."
     )
+    # Remove references without mapped reads
+    # alns_in_ref = defaultdict(int)
+    # for aln in tqdm.tqdm(
+    #     samfile.fetch(until_eof=True),
+    #     total=samfile.mapped,
+    #     leave=False,
+    #     ncols=80,
+    #     desc=f"Alignments processed",
+    # ):
+    #     alns_in_ref[aln.reference_name] += 1
+
+    if plot:
+        # Check if image folder exists
+        if not os.path.exists(plots_dir):
+            os.makedirs(plots_dir)
 
     references = [
         chrom.contig
@@ -442,9 +601,7 @@ def process_bam(bam, threads=1, reference_lengths=None, scale=1e6, min_read_coun
         sys.exit(1)
 
     logging.info(f"Keeping {len(references):,} references")
-    
     params = zip([bam] * len(references), references)
-
     try:
         logging.info(f"Getting stats for each reference...")
 
@@ -452,7 +609,12 @@ def process_bam(bam, threads=1, reference_lengths=None, scale=1e6, min_read_coun
             data = list(
                 map(
                     functools.partial(
-                        get_bam_stats, ref_lengths=ref_lengths, scale=scale
+                        get_bam_stats,
+                        ref_lengths=ref_lengths,
+                        min_read_ani=min_read_ani,
+                        scale=scale,
+                        plot=plot,
+                        plots_dir=plots_dir,
                     ),
                     params,
                 )
@@ -464,17 +626,27 @@ def process_bam(bam, threads=1, reference_lengths=None, scale=1e6, min_read_coun
                 initializer=initializer,
                 initargs=([params, ref_lengths, scale],),
             )
-            c_size = calc_chunksize(
-                n_workers=threads, len_iterable=len(references), factor=4
-            )
+
+            if chunksize is not None:
+                c_size = chunksize
+            else:
+                c_size = calc_chunksize(
+                    n_workers=threads, len_iterable=len(references), factor=4
+                )
+
             data = list(
                 tqdm.tqdm(
                     p.imap_unordered(
                         functools.partial(
-                            get_bam_stats, ref_lengths=ref_lengths, scale=scale
+                            get_bam_stats,
+                            ref_lengths=ref_lengths,
+                            min_read_ani=min_read_ani,
+                            scale=scale,
+                            plot=plot,
+                            plots_dir=plots_dir,
                         ),
                         params,
-                        chunksize=20,
+                        chunksize=c_size,
                     ),
                     total=len(references),
                     leave=False,
@@ -514,15 +686,17 @@ def filter_reference_BAM(
     """
     logging.info("Filtering stats...")
     logging.info(
-        f"min_read_count >= {filter_conditions['min_read_count']} & min_read_length >= {filter_conditions['min_read_length']} & min_read_ani >= {filter_conditions['min_read_ani']} & min_expected_breadth_ratio >= {filter_conditions['min_expected_breadth_ratio']} &  min_breadth >= {filter_conditions['min_breadth']} & min_coverage_evenness >= {filter_conditions['min_coverage_evenness']}"
+        f"min_read_count >= {filter_conditions['min_read_count']} & min_read_length >= {filter_conditions['min_read_length']} & min_avg_read_ani >= {filter_conditions['min_avg_read_ani']} & min_expected_breadth_ratio >= {filter_conditions['min_expected_breadth_ratio']} &  min_breadth >= {filter_conditions['min_breadth']} & min_coverage_evenness >= {filter_conditions['min_coverage_evenness']} & min_norm_entropy >= {filter_conditions['min_norm_entropy']} & min_norm_gini <= {filter_conditions['min_norm_gini']}"
     )
     df_filtered = df.loc[
         (df["n_reads"] >= filter_conditions["min_read_count"])
         & (df["read_length_mean"] >= filter_conditions["min_read_length"])
-        & (df["read_ani_mean"] >= filter_conditions["min_read_ani"])
+        & (df["read_ani_mean"] >= filter_conditions["min_avg_read_ani"])
         & (df["breadth_exp_ratio"] >= filter_conditions["min_expected_breadth_ratio"])
         & (df["breadth"] >= filter_conditions["min_breadth"])
         & (df["cov_evenness"] >= filter_conditions["min_coverage_evenness"])
+        & (df["norm_entropy"] >= filter_conditions["min_norm_entropy"])
+        & (df["norm_gini"] <= filter_conditions["min_norm_gini"])
     ]
     if len(df_filtered.index) > 0:
         logging.info(f"Saving filtered stats...")
